*********************************************************************
*********************************************************************
**  Integrate the Quasi Geostrophic potential vorticity equation.  **
*********************************************************************
*********************************************************************

!     Equation represents ocean circulation driven by a specified wind
!     stress, showing in a time average two gyres with a "Gulf Stream"
!     on the left side of the domain:
!      
!     \partial_t q + r J(\psi,q) + \beta psi_x =
!        -rkb\zeta + rkh\nabla^2\zeta - rkh2\nabla^4\zeta + (\tau_0/f\rho_0)curl\tau
!
!     where
!      
!     q        = \zeta-Froud\psi   is the potential vorticity
!     \zeta    = \nabla^2\psi      is the relative corticity
!     J(\psi,q)= \psi_y q_x - \psi_x q_y = Jacobian, nonlinear advection term
!     r        = multiplying the nonlinear advection term.  Sort of a Rossby number.
!     rkb      = bottom friction
!     rkh      = horizontal friction
!     rkh2     = biharmonic horizontal friction
!     Froud    = Froud number
!     \tau_0 curl\tau = specified wind stress curl:  the forcing function.
!     \beta    = gradient of coriolis parameter with respect to latitude
!     f        = coriolis parameter

!     Input paramters are explained in subroutine init.
      
*     April 2002: changed time stepping scheme from euler forward to
*     leap from+robert filter.  Code clean-up.  Added free surface
*     (vertical stretching) term and therefore had to change from
*     Poisson to Helmholtz solver.

      implicit none
      include "parameters.h"
      integer i,j,np,mp,nt,mrefp,nx1p,ny1p
     &     ,knt,ncyc,nu1,nu2
     &     ,jbeg,jend,ibeg,iend,ncyc_test
      real*8 pi,h1,c1,c2,c3,c4,c5,c6,vort_max
      parameter(pi=3.14159265)
C     Arrays:
      real*8 psiold(n,m),psinow(n,m),psinew(n,m),psimean(n,m)
     &     ,vortold(n,m),vortnow(n,m),vortnew(n,m),vortmean(n,m)
     &     ,zetold(n,m),zetnow(n,m),zetnew(n,m)
     &     ,dl2zetold(n,m),dl4zetold(n,m),adv_vort(n,m),curlt(n,m)
      logical test_helmholtz, calc_mean


c     Set parameters and forcing:
      CALL INIT(CURLT)

c     Read restart if exists:
      read(36,end=999,err=999) psiold,psinow,psinew
     &     ,zetold,zetnow,zetnew
     &     ,vortold,vortnow,vortnew
      write(*,*) ' Found restart, continuing from it.'
      goto 1000
 999  continue
      write(*,*) ' No restart, initializing to zero:'
c     Initialize to zero if no restart:
      do i=1,n
        do j=1,m
          psiold(i,j)=0.0
          psinow(i,j)=0.0
          zetold(i,j)=0.0
          zetnow(i,j)=0.0
          vortold(i,j)=0.0 ! exp(-(i-N/2.)**2/5.**2-(j-M/2.)**2/5.**2)
          vortnow(i,j)=vortold(i,j)
          psimean(i,j)=0.0
          vortmean(i,j)=0.0
       end do
      end do
 1000 continue

!     Some parameters for the multigrid solver:
!     This seems to assume dx=dy:
      H1=Lx/NX1
      NCYC=3
      NU1=1
      NU2=2
C   IN ORDER NOT TO USE CONSTANTS IN THE CALL TO POISS1:
      MREFP=MREFIN
      NX1P=NX1
      NY1P=NY1
      NP=N
      MP=M
      KNT=NPRINT/5

      IBEG=2
      IEND=N-1
      JBEG=2
      JEND=M-1

      IF (KNT.EQ.0) KNT = 1
      WRITE(6,*) ' N, M =',N,M
      WRITE(6,*) 'DX,DY,DT=',DX,DY,DT


!     Test Helmholtz solver:
      test_helmholtz=.false.
      if (test_helmholtz) then
        write(*,*) "Testing helmholtz solver:"
        do i=ibeg,iend
          do j=jbeg,jend
            psinow(i,j)=exp(-(i-N/2.)**2/5.**2-(j-M/2.)**2/5.**2)
          end do
        end do
        call del2(psinow,zetnow,n,m,dx,dy)
        do i=ibeg,iend
          do j=jbeg,jend
            vortnow(i,j)=zetnow(i,j)-Froud*psinow(i,j)
          end do
        end do
        ncyc_test=2
!       Use vortnew as initial guess (vortnew is zero here)
        call helmholt(vortnew,vortnow,psinew,froud,nx1p,ny1p
     &       ,mrefp,h1,nu1,nu2,ncyc_test,np,mp)
        call plotrs(psinew,n,m,'steam function:')
        do i=ibeg,iend
          do j=jbeg,jend
            psinow(i,j)=psinew(i,j)-psinow(i,j)
          end do
        end do
        call plotrs(psinow,n,m,'steam function:')
        write(*,*) "End of testing."
        stop
      end if

!     Define some constants in order to save operations later:
      C1=2*DT*beta/(2.0*DX)
      C2=2*DT*R
      C3=2*DT*RKB
      C4=2*DT*RKH
      C5=2*DT*RKH2
      C6=2*DT

!     time stepping loop starts here:
      DO NT=NBEG,NEND
        IF ( (NT.EQ.1.OR.NT/KNT*KNT.EQ.NT)
     &       .AND..NOT.(NPRINT.EQ.999) ) then
          write(6,*)' Time step=',NT
          call flush(6)
        end if
c       Check if the model bombed:
        vort_max=0.0
        do i=ibeg,iend
          do j=jbeg,jend
            vort_max=max(vort_max,abs(vortnew(i,j)))
          end do
        end do
        if (vort_max.gt.1.e25) then
          write(*,*) " --->>> Oops, model is bombing..."
          stop
        end if
        
c       Calculate nonlinear advection term (Arakawa discretization of
c       the Jacobian):
        call arajac(psinow,vortnow,adv_vort,n,m,dx,dy)

c       Calculate \nabla^2\zeta & \nabla^4\zeta to be used for friction term:
        call del2(zetold,dl2zetold,n,m,dx,dy)
        call del2(dl2zetold,dl4zetold,n,m,dx,dy)

c       Step in time:
        DO I=IBEG,IEND
          DO J=JBEG,JEND
            VORTNEW(I,J)=VORTOLD(I,J)+(
     &           -C1*(PSINOW(I+1,J)-PSINOW(I-1,J))
     &           -C2*ADV_VORT(I,J)
     &           -C3*ZETNOW(I,J)
     &           +C4*DL2ZETOLD(I,J)
     &           -C5*DL4ZETOLD(I,J)
     &           +c6*CURLT(I,J)
     &           )
          end do
        end do

!       Solve for stream function from vortivity:        
!        CALL POISS1 (PSINOW,VORTNEW,PSINEW,       NX1P,NY1P
!     &      ,MREFP,H1,NU1,NU2,NCYC,NP,MP)
        call HELMHOLT(PSINOW,VORTNEW,PSINEW,Froud,NX1P,NY1P
     &       ,MREFP,H1,NU1,NU2,NCYC,NP,MP)

!       calculate new relative vorticity:
        call del2(psinew,zetnew,n,m,dx,dy)

c       Robert time filter:
        DO I=IBEG,IEND
          DO J=JBEG,JEND
            vortnow(i,j)=vortnow(i,j) + robert_filter
     &           *(vortnew(i,j)-2*vortnow(i,j)+vortold(i,j))
          end do
        end do

c       Prepare for next time step:
        do i=ibeg,iend
          do j=jbeg,jend
            vortold(i,j)=vortnow(i,j)
            vortnow(i,j)=vortnew(i,j)
            psiold(i,j)=psinow(i,j)
            psinow(i,j)=psinew(i,j)
            zetold(i,j)=zetnow(i,j)
            zetnow(i,j)=zetnew(i,j)
          end do
        end do

c       Write results to output file for plotting:
        IF (NT/NPRINT*NPRINT.EQ.NT) THEN
          CALL PLOTRS(PSINEW,N,M,'steam function:')
          CALL PLOTRS(vortNEW,N,M,'potential vorticity:')
!          CALL PLOTRS(adv_vort,N,M,'jacobian:')
          WRITE(6,*)' ... '
        END IF

!       Calc time-mean fields
        calc_mean=.true.
        if (calc_mean) then
          do i=ibeg,iend
            do j=jbeg,jend
              vortmean(i,j)=vortmean(i,j)+vortnew(i,j)/(nend-nbeg+1)
              psimean(i,j) =psimean(i,j) +psinew(i,j) /(nend-nbeg+1)
            end do
          end do
        end if
      end do

!     Plot time-mean fields
      if (calc_mean) then
        call plotrs(psimean,n,m,'steam function:')
        call plotrs(vortmean,n,m,'potential vorticity:')
      end if

c     Write restart file:
      rewind(36)
      write(36) psiold,psinow,psinew
     &     ,zetold,zetnow,zetnew
     &     ,vortold,vortnow,vortnew

      stop
      end


      SUBROUTINE INIT(CURLT)
*********************************************************************
*********************************************************************
**   Initializes parameters.                                       **
*********************************************************************
*********************************************************************

      IMPLICIT none
      include "parameters.h"
      integer i,j
      real*8 OBSERV(N,M)
      real*8 CURLT(N,M)
      real*8 pi,EPSDT
      DATA PI/3.14159265/

      namelist /nl_input/
     &     NBEG,NEND                              ! beginning and end time steps,
     &     ,NPRINT                                ! frequency of requested printout.
     &     ,EPSDT                                 ! time step
     &     ,RKB,RKH,RKH2                          ! bottom, horizintal and biharmonic
                                                  !         horizontal friction coefficients
     &     ,beta                                  ! gradient of coriolis force ert to latitude
     &     ,Rossby_radius                         ! roughly: scale of eddies
     &     ,R                                     ! factor infront of nonlinear advection term
     &     ,Lx,Ly                                 ! domain size
     &     ,tau0                                  ! forcing amplitude
     &     ,coriolis                              ! coriolis force
     &     ,rho0                                  ! reference constant water density
     &     ,robert_filter                         ! used to fix leap frog instability
      

C     EPSDT IS THE FACTOR MULTIPLYING THE TIME STEP BELOW:
      open(15,file='input.data',form='formatted')
      open(26,file='Output/output.dat',form='formatted')
      open(36,file='Output/restart.dat',form='unformatted')
      read(15,nl_input)
!     write(6,nl_input)
      
      DX=Lx/(N-1)
      DY=Ly/(M-1)

      DT=EPSDT ! *1./(2.*RKH*(1./DX**2+DX**2))

      Froud=1.0/Rossby_radius**2

      do i=1,n
        do j=1,m
          CURLT(I,J)=-(2.0*pi*tau0/(Lx*rho0*coriolis))
     &         *sin(2.0*pi*(j-1)/(m-1))
        end do
      end do

      return
      end


      SUBROUTINE ARAJAC(PSI,Q,RJACOBIAN,M,N,dx,dy)
*********************************************************************
*********************************************************************
**   Calculates the arakawa jacobian J(psi,Q)=R                    **
*********************************************************************
*********************************************************************
      IMPLICIT none
      integer i,j,n,m,m1,n1
      real*8 PSI(M,N),Q(M,N),RJACOBIAN(M,N),dx,dy

      M1=M-1
      N1=N-1
      DO I=2,M1
        DO J=2,N1
          RJACOBIAN(I,J)=((PSI(I-1,J)-PSI(I,J-1))*Q(I-1,J-1)+
     &     (PSI(I-1,J-1)+PSI(I-1,J)-PSI(I+1,J-1)-PSI(I+1,J))*Q(I,J-1)+
     &     (PSI(I,J-1)-PSI(I+1,J))*Q(I+1,J-1)+
     &     (PSI(I-1,J+1)+PSI(I,J+1)-PSI(I-1,J-1)-PSI(I,J-1))*Q(I-1,J)+
     &     (PSI(I,J-1)+PSI(I+1,J-1)-PSI(I,J+1)-PSI(I+1,J+1))*Q(I+1,J)+
     &     (PSI(I,J+1)-PSI(I-1,J))*Q(I-1,J+1)+
     &     (PSI(I+1,J)+PSI(I+1,J+1)-PSI(I-1,J)-PSI(I-1,J+1))*Q(I,J+1)+
     &     (PSI(I+1,J)-PSI(I,J+1))*Q(I+1,J+1))/(12.*DX*DY)
        end do
      end do

      return
      end


      SUBROUTINE DEL2(ARRY,DL2ARY,N,M,dx,dy)
*********************************************************************
*********************************************************************
**   Calculates     \nabla^2 ARRY(x,y), returns result in dl2ary.  **
** dimension of arrays are (N,M).                                  **
*********************************************************************
*********************************************************************

      implicit none
      integer n,m,i,j
      real*8 ARRY(N,M),DL2ARY(N,M),dx,dy

      DO I=2,N-1
        DO J=2,M-1
          DL2ARY(I,J)=(ARRY(I+1,J)-2.0*ARRY(I,J)+ARRY(I-1,J))/DX**2
     &         +(ARRY(I,J+1)-2.0*ARRY(I,J)+ARRY(I,J-1))/DY**2
        end do
      end do

      return
      end


      SUBROUTINE PLOTRS(ARRAY,N,M,TITLE)
**********************************************************************
**********************************************************************
**  Write data for contour from the input array SURFCE(N,M).                 **
**********************************************************************
**********************************************************************
      implicit none
      integer   i,  j, n,m
      real*8 array(n,m)
      character *(*) title
      character*32 fmt
      character*1 quote
      
!     write(26,*) title
      if (m.le.99) then
        write(fmt,'("(",i2,"g14.6)")') m
      elseif (m.le.999) then
        write(fmt,'("(",i3,"g14.6)")') m
      endif
      write(26,fmt) ((array(i,j),j=1,m),i=1,n)
      call flush(26)
      
      return
      end
